@@ -124,6 +124,7 @@ void pci_register_io_region(PCIDevice *pci_dev, int region_num,
                             PCIMapIORegionFunc *map_func)
 {
     PCIIORegion *r;
+    uint32_t addr;
 
     if ((unsigned int)region_num >= PCI_NUM_REGIONS)
         return;
@@ -132,6 +133,12 @@ void pci_register_io_region(PCIDevice *pci_dev, int region_num,
     r->size = size;
     r->type = type;
     r->map_func = map_func;
+    if (region_num == PCI_ROM_SLOT) {
+        addr = 0x30;
+    } else {
+        addr = 0x10 + region_num * 4;
+    }
+    *(uint32_t *)(pci_dev->config + addr) = cpu_to_le32(type);
 }
 
 static void pci_addr_writel(void* opaque, uint32_t addr, uint32_t val)
@@ -353,9 +360,6 @@ static void pci_data_write(void *opaque, uint32_t addr,
     if (!(s->config_reg & (1 << 31))) {
         return;
     }
-    if ((s->config_reg & 0x3) != 0) {
-        return;
-    }
     bus_num = (s->config_reg >> 16) & 0xff;
     if (bus_num != 0)
         return;
@@ -380,8 +384,6 @@ static uint32_t pci_data_read(void *opaque, uint32_t addr,
 
     if (!(s->config_reg & (1 << 31)))
         goto fail;
-    if ((s->config_reg & 0x3) != 0)
-        goto fail;
     bus_num = (s->config_reg >> 16) & 0xff;
     if (bus_num != 0)
         goto fail;
@@ -1837,3 +1839,17 @@ void pci_bios_init(void)
         }
     }
 }
+
+/* Initialize a PCI NIC.  */
+void pci_nic_init(PCIBus *bus, NICInfo *nd)
+{
+    if (strcmp(nd->model, "ne2k_pci") == 0) {
+        pci_ne2000_init(bus, nd);
+    } else if (strcmp(nd->model, "rtl8139") == 0) {
+        pci_rtl8139_init(bus, nd);
+    } else {
+        fprintf(stderr, "qemu: Unsupported NIC: %s\n", nd->model);
+        exit (1);
+    }
+}
+
